#ifndef UU_NETWORKS_IMPL_STORES_MLORDEREDLAYERSTORE_H_
#define UU_NETWORKS_IMPL_STORES_MLORDEREDLAYERSTORE_H_

#include "networks/_impl/stores/VertexOverlappingOrderedLayerStore.hpp"
#include "networks/Network.hpp"

namespace uu {
namespace net {

typedef VertexOverlappingOrderedLayerStore<Network> MLOrderedLayerStore;

}
}

#endif
