#ifndef UU_NETWORKS_IMPL_STORES_MLLAYERSTORE_H_
#define UU_NETWORKS_IMPL_STORES_MLLAYERSTORE_H_

#include "networks/_impl/stores/VertexOverlappingLayerStore.hpp"
#include "networks/Network.hpp"

namespace uu {
namespace net {

typedef VertexOverlappingLayerStore<Network> MLLayerStore;

}
}

#endif
