// This file is part of Eigen, a lightweight C++ template library
// for linear algebra.
//
// Copyright (C) 2014 Benoit Steiner <benoit.steiner.goog@gmail.com>
//
// This Source Code Form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

#ifndef EIGEN_CXX11_TENSOR_TENSOR_SHUFFLING_H
#define EIGEN_CXX11_TENSOR_TENSOR_SHUFFLING_H

namespace Eigen {

/** \class TensorShuffling
  * \ingroup CXX11_Tensor_Module
  *
  * \brief Tensor shuffling class.
  *
  *
  */
namespace internal {
template<typename Shuffle, typename XprType>
struct traits<TensorShufflingOp<Shuffle, XprType> > : public traits<XprType>
{
    typedef typename XprType::Scalar Scalar;
    typedef traits<XprType> XprTraits;
    typedef typename XprTraits::StorageKind StorageKind;
    typedef typename XprTraits::Index Index;
    typedef typename XprType::Nested Nested;
    typedef typename remove_reference<Nested>::type _Nested;
    static const int NumDimensions = XprTraits::NumDimensions;
    static const int Layout = XprTraits::Layout;
};

template<typename Shuffle, typename XprType>
struct eval<TensorShufflingOp<Shuffle, XprType>, Eigen::Dense>
{
    typedef const TensorShufflingOp<Shuffle, XprType>& type;
};

template<typename Shuffle, typename XprType>
struct nested<TensorShufflingOp<Shuffle, XprType>, 1, typename eval<TensorShufflingOp<Shuffle, XprType> >::type>
{
    typedef TensorShufflingOp<Shuffle, XprType> type;
};

}  // end namespace internal



template<typename Shuffle, typename XprType>
class TensorShufflingOp : public TensorBase<TensorShufflingOp<Shuffle, XprType> >
{
  public:
    typedef typename Eigen::internal::traits<TensorShufflingOp>::Scalar Scalar;
    typedef typename Eigen::NumTraits<Scalar>::Real RealScalar;
    typedef typename XprType::CoeffReturnType CoeffReturnType;
    typedef typename Eigen::internal::nested<TensorShufflingOp>::type Nested;
    typedef typename Eigen::internal::traits<TensorShufflingOp>::StorageKind StorageKind;
    typedef typename Eigen::internal::traits<TensorShufflingOp>::Index Index;

    EIGEN_DEVICE_FUNC EIGEN_STRONG_INLINE
    TensorShufflingOp(const XprType& expr, const Shuffle& shuffle)
        : m_xpr(expr), m_shuffle(shuffle) {}

    EIGEN_DEVICE_FUNC
    const Shuffle&
    shufflePermutation() const
    {
        return m_shuffle;
    }

    EIGEN_DEVICE_FUNC
    const typename internal::remove_all<typename XprType::Nested>::type&
    expression() const
    {
        return m_xpr;
    }

    EIGEN_DEVICE_FUNC
    EIGEN_STRONG_INLINE TensorShufflingOp&
    operator = (const TensorShufflingOp& other)
    {
        typedef TensorAssignOp<TensorShufflingOp, const TensorShufflingOp> Assign;
        Assign assign(*this, other);
        internal::TensorExecutor<const Assign, DefaultDevice>::run(assign, DefaultDevice());
        return *this;
    }

    template<typename OtherDerived>
    EIGEN_DEVICE_FUNC
    EIGEN_STRONG_INLINE TensorShufflingOp&
    operator = (const OtherDerived& other)
    {
        typedef TensorAssignOp<TensorShufflingOp, const OtherDerived> Assign;
        Assign assign(*this, other);
        internal::TensorExecutor<const Assign, DefaultDevice>::run(assign, DefaultDevice());
        return *this;
    }

  protected:
    typename XprType::Nested m_xpr;
    const Shuffle m_shuffle;
};


// Eval as rvalue
template<typename Shuffle, typename ArgType, typename Device>
struct TensorEvaluator<const TensorShufflingOp<Shuffle, ArgType>, Device>
{
    typedef TensorShufflingOp<Shuffle, ArgType> XprType;
    typedef typename XprType::Index Index;
    static const int NumDims = internal::array_size<typename TensorEvaluator<ArgType, Device>::Dimensions>::value;
    typedef DSizes<Index, NumDims> Dimensions;
    typedef typename XprType::Scalar Scalar;
    typedef typename XprType::CoeffReturnType CoeffReturnType;
    typedef typename PacketType<CoeffReturnType, Device>::type PacketReturnType;
    static const int PacketSize = internal::unpacket_traits<PacketReturnType>::size;

    enum
    {
        IsAligned = false,
        PacketAccess = (internal::packet_traits<Scalar>::size > 1),
        Layout = TensorEvaluator<ArgType, Device>::Layout,
        CoordAccess = false,  // to be implemented
        RawAccess = false
    };

    EIGEN_DEVICE_FUNC EIGEN_STRONG_INLINE
    TensorEvaluator(const XprType& op, const Device& device)
        : m_impl(op.expression(), device)
    {
        const typename TensorEvaluator<ArgType, Device>::Dimensions& input_dims = m_impl.dimensions();
        const Shuffle& shuffle = op.shufflePermutation();

        for (int i = 0; i < NumDims; ++i)
        {
            m_dimensions[i] = input_dims[shuffle[i]];
        }

        array<Index, NumDims> inputStrides;

        if (static_cast<int>(Layout) == static_cast<int>(ColMajor))
        {
            inputStrides[0] = 1;
            m_outputStrides[0] = 1;

            for (int i = 1; i < NumDims; ++i)
            {
                inputStrides[i] = inputStrides[i - 1] * input_dims[i - 1];
                m_outputStrides[i] = m_outputStrides[i - 1] * m_dimensions[i - 1];
            }
        }

        else
        {
            inputStrides[NumDims - 1] = 1;
            m_outputStrides[NumDims - 1] = 1;

            for (int i = NumDims - 2; i >= 0; --i)
            {
                inputStrides[i] = inputStrides[i + 1] * input_dims[i + 1];
                m_outputStrides[i] = m_outputStrides[i + 1] * m_dimensions[i + 1];
            }
        }

        for (int i = 0; i < NumDims; ++i)
        {
            m_inputStrides[i] = inputStrides[shuffle[i]];
        }
    }

    EIGEN_DEVICE_FUNC EIGEN_STRONG_INLINE const Dimensions&
    dimensions() const
    {
        return m_dimensions;
    }

    EIGEN_DEVICE_FUNC EIGEN_STRONG_INLINE bool
    evalSubExprsIfNeeded(Scalar* /*data*/)
    {
        m_impl.evalSubExprsIfNeeded(NULL);
        return true;
    }
    EIGEN_DEVICE_FUNC EIGEN_STRONG_INLINE void
    cleanup()
    {
        m_impl.cleanup();
    }

    EIGEN_DEVICE_FUNC EIGEN_STRONG_INLINE CoeffReturnType
    coeff(Index index) const
    {
        return m_impl.coeff(srcCoeff(index));
    }

    template<int LoadMode>
    EIGEN_DEVICE_FUNC EIGEN_STRONG_INLINE PacketReturnType
    packet(Index index) const
    {
        EIGEN_STATIC_ASSERT((PacketSize > 1), YOU_MADE_A_PROGRAMMING_MISTAKE)
        eigen_assert(index+PacketSize-1 < dimensions().TotalSize());

        EIGEN_ALIGN_MAX typename internal::remove_const<CoeffReturnType>::type values[PacketSize];

        for (int i = 0; i < PacketSize; ++i)
        {
            values[i] = coeff(index+i);
        }

        PacketReturnType rslt = internal::pload<PacketReturnType>(values);
        return rslt;
    }

    EIGEN_DEVICE_FUNC EIGEN_STRONG_INLINE TensorOpCost
    costPerCoeff(bool vectorized) const
    {
        const double compute_cost = NumDims * (2 * TensorOpCost::AddCost<Index>() +
                                               2 * TensorOpCost::MulCost<Index>() +
                                               TensorOpCost::DivCost<Index>());
        return m_impl.costPerCoeff(vectorized) +
               TensorOpCost(0, 0, compute_cost, false /* vectorized */, PacketSize);
    }

    EIGEN_DEVICE_FUNC Scalar*
    data() const
    {
        return NULL;
    }

  protected:
    EIGEN_DEVICE_FUNC EIGEN_STRONG_INLINE Index
    srcCoeff(Index index) const
    {
        Index inputIndex = 0;

        if (static_cast<int>(Layout) == static_cast<int>(ColMajor))
        {
            for (int i = NumDims - 1; i > 0; --i)
            {
                const Index idx = index / m_outputStrides[i];
                inputIndex += idx * m_inputStrides[i];
                index -= idx * m_outputStrides[i];
            }

            return inputIndex + index * m_inputStrides[0];
        }

        else
        {
            for (int i = 0; i < NumDims - 1; ++i)
            {
                const Index idx = index / m_outputStrides[i];
                inputIndex += idx * m_inputStrides[i];
                index -= idx * m_outputStrides[i];
            }

            return inputIndex + index * m_inputStrides[NumDims - 1];
        }
    }

    Dimensions m_dimensions;
    array<Index, NumDims> m_outputStrides;
    array<Index, NumDims> m_inputStrides;
    TensorEvaluator<ArgType, Device> m_impl;
};


// Eval as lvalue
template<typename Shuffle, typename ArgType, typename Device>
struct TensorEvaluator<TensorShufflingOp<Shuffle, ArgType>, Device>
    : public TensorEvaluator<const TensorShufflingOp<Shuffle, ArgType>, Device>
{
    typedef TensorEvaluator<const TensorShufflingOp<Shuffle, ArgType>, Device> Base;

    typedef TensorShufflingOp<Shuffle, ArgType> XprType;
    typedef typename XprType::Index Index;
    static const int NumDims = internal::array_size<typename TensorEvaluator<ArgType, Device>::Dimensions>::value;
    typedef DSizes<Index, NumDims> Dimensions;
    typedef typename XprType::Scalar Scalar;
    typedef typename XprType::CoeffReturnType CoeffReturnType;
    typedef typename PacketType<CoeffReturnType, Device>::type PacketReturnType;
    static const int PacketSize = internal::unpacket_traits<PacketReturnType>::size;

    enum
    {
        IsAligned = false,
        PacketAccess = (internal::packet_traits<Scalar>::size > 1),
        RawAccess = false
    };

    EIGEN_DEVICE_FUNC EIGEN_STRONG_INLINE
    TensorEvaluator(const XprType& op, const Device& device)
        : Base(op, device)
    { }

    EIGEN_DEVICE_FUNC EIGEN_STRONG_INLINE CoeffReturnType&
    coeffRef(Index index)
    {
        return this->m_impl.coeffRef(this->srcCoeff(index));
    }

    template <int StoreMode> EIGEN_STRONG_INLINE
    void
    writePacket(Index index, const PacketReturnType& x)
    {
        EIGEN_STATIC_ASSERT((PacketSize > 1), YOU_MADE_A_PROGRAMMING_MISTAKE)

        EIGEN_ALIGN_MAX typename internal::remove_const<CoeffReturnType>::type values[PacketSize];
        internal::pstore<CoeffReturnType, PacketReturnType>(values, x);

        for (int i = 0; i < PacketSize; ++i)
        {
            this->coeffRef(index+i) = values[i];
        }
    }
};


} // end namespace Eigen

#endif // EIGEN_CXX11_TENSOR_TENSOR_SHUFFLING_H
